#!/usr/bin/env python

from .entrypoints import EntryPoint
from .link import Link

def __clusterAvailability(entryPoints, virtualMachines):
	if not virtualMachines:
		return 1.0

	cluster = virtualMachines[0].machine.cluster
	assert(vm.machine.cluster == cluster for vm in virtualMachines)
	
	networkAvailability = 1.0 - reduce(lambda p,l: p * (1.0 - l.availability), filter(lambda l: l.destination in entryPoints, cluster.accessLinks.values()), 1.0)
	machineAvailability = 1.0 - reduce(lambda p,m: p * (1.0 - m.availability), virtualMachines, 1.0)
	
	return networkAvailability * cluster.availability * machineAvailability

def computeAvailability(entryPoints, masterVM):
	if not masterVM or not entryPoints:
		return 0.0
	assert(masterVM.isMaster)
	
	virtualMachines = masterVM.instances
	assert(all(vm in vm.machine.virtualMachines for vm in virtualMachines ))
	masterCluster = masterVM.machine.cluster
	assert(not vm.isMaster()or vm == masterVM for vm in virtualMachines)
	
	clusters = {}
	for vm in virtualMachines:
		clusters.setdefault(vm.machine.cluster, []).append(vm)
	
	a = 1.0
	for c, vms in clusters.iteritems(): 
		a *= 1.0 - __clusterAvailability(entryPoints, vms) * (1.0 if c == masterCluster else masterCluster.synchronizationLinks[c.name].availability) 

	return 1.0 - a
	
class VirtualMachine:
	
	def __init__(self, machine, function, capacity):
		self.__function = function
		self.__requestAssignments = {}
		self.__capacity = float(capacity)
		self.__master = None
		self.__copies = [self]
		self.__machine = machine
	
	@property
	def instances(self):
		return self.__copies
	
	@property
	def volume(self):
		return reduce(lambda s, d: s + d,self.__requestAssignments.values(),0.0)

	@property
	def capacity(self):
		return self.__capacity

	@property
	def residualCapacity(self):
		return self.__capacity - self.volume

	@property
	def master(self):
		return self.__master
	
	def isMaster(self):
		return self.__master == None
	
	@property
	def requests(self):
		return self.__requestAssignments.keys()
	
	@property
	def function(self):
		return self.__function
	
	@property
	def machine(self):
		return self.__machine
	
	def backup(self, machine):
		assert(self.isMaster())
		assert(self.__machine != machine)
		backup = VirtualMachine(machine, self.__function, self.__capacity)
		backup.__requestAssignments = self.__requestAssignments
		backup.__master = self
		
		self.__copies.append(backup)
		backup.__copies = self.__copies
		machine.addVM(backup)
		return backup
	
	def move(self, newMachine):
		assert(not self.isMaster() or not newMachine.hasMasterVirtualMachine(self.__function))
		assert(newMachine.isAvailable(self.volume))
		self.__machine = newMachine
	
	def addRequest(self, request, assignedDemand):
		assert(request.function == self.__function)
		assert(self.machine.volume + assignedDemand <= self.machine.capacity)
		self.__requestAssignments.setdefault(request, 0.0)
		self.__requestAssignments[request] += assignedDemand

	def hasRequest(self, request):
		return self.__requestAssignments.has_key(request)

	def assignment(self, request):
		return self.__requestAssignments[request]

	@property
	def availability(self):
		return self.__function.availability * self.__machine.availability

	def changeMaster(self, virtualMachine):
		assert(self.isMaster())
		assert(virtualMachine in self.__copies)
		for vm in self.__copies: 
			vm.__master = virtualMachine
		virtualMachine.__master = None

	def removeRequest(self, request, demand):
		assert(self.isMaster())
		assert(self.__requestAssignments.has_key(request))
		request.restore(demand, self)
		if demand == self.__requestAssignments[request]:
			self.__requestAssignments.pop(request)
		else:
			self.__requestAssignments[request] -= float(demand)

	def __str__(self):
		return "<"+str(self.__machine)+", "+str(self.__function)+", ("+str(self.__master)+")>"
	
	def __repr__(self):
		return "<"+self.__machine.name+", "+self.__function.name+">"

class Machine(object):
	
	LOAD_BALANCING = False
	__counter = 0
	VM_SIZE_FACTOR = 1.0
	
	def __init__(self, availability = 1.0, capacity = 200.0, cluster  = None, objid = None):
		if objid:
			assert Machine.__counter <= int(objid)
			Machine.__counter = int(objid)
		else:
			Machine.__counter += 1
		self.__name = "m"+str(Machine.__counter)
		self.__capacity = float(capacity)
		self.cluster = cluster
		self.__virtualMachines = {}
		self.__availability = float(availability)
	
	@property
	def virtualMachines(self):
		return sum(self.__virtualMachines.values(), [])
	
	def masterVirtualMachine(self, function):
		return self.__checkMaster(function, False)
	
	def hasMasterVirtualMachine(self, function):
		return self.__checkMaster(function, False) != None
	
	@property
	def requests(self):
		return [r for vm in self.virtualMachines if vm.isMaster() for r in vm.requests]
	
	@property
	def name(self):
		return self.__name
	
	@property
	def capacity(self):
		return self.__capacity
	
	@property
	def volume(self):
		return reduce(lambda s, vm: s + vm.volume, sum(self.__virtualMachines.values(),[]), 0.0)
	
	@property
	def availability(self):
		return self.__availability
	
	def addRequest(self, request, assignedDemand):
		assert(assignedDemand > 0)
		assert(self.volume + assignedDemand <= self.__capacity)
		master = self.__checkMaster(request.function, True)
		master.addRequest(request, assignedDemand)
		request.serve(assignedDemand, master)
	
	def removeRequest(self, request, assignedDemand):
		master = self.__checkMaster(request.function, False)
		assert(master != None)
		master.removeRequest(request, assignedDemand)
	
	def hasRequest(self, request):
		master = self.__checkMaster(request.function, False)
		return master and master.hasRequest(request)
	
	def assignment(self, request):
		master = self.__checkMaster(request.function, False)
		assert(master != None)
		return master.assignment(request)
								
	def __checkMaster(self, function, create = False):
		masters = filter(lambda vm: vm.isMaster(), self.__virtualMachines.setdefault(function.name, []))
		assert(len(masters) <= 1)
		
		if not masters and create:
			masters.append(VirtualMachine(self, function, self.__capacity * Machine.VM_SIZE_FACTOR))
			self.__virtualMachines[function.name] += masters

		return masters[0] if create or masters else None  
		
	
	def addVM(self, virtualMachine):
		self.__virtualMachines.setdefault(virtualMachine.function.name, [])
		assert(not virtualMachine.isMaster() or not any(vm.isMaster() for vm in self.__virtualMachines[virtualMachine.function.name]))
		assert(not any(virtualMachine.master == vm for vm in self.__virtualMachines[virtualMachine.function.name]))
		assert(not any(not vm.isMaster() and virtualMachine.master == vm.master for vm in self.__virtualMachines[virtualMachine.function.name]))
		assert(self.volume + virtualMachine.volume <= self.__capacity)
		assert(not any(vm in self.virtualMachines for vm in virtualMachine.instances))
		self.__virtualMachines[virtualMachine.function.name].append(virtualMachine)
		
	def residualCapacity(self, function = None):
		master = self.__checkMaster(function, False) if function else None
		cap = min(self.__capacity - self.volume, master.residualCapacity if master else self.__capacity)
		assert cap >= 0, "residualCapacity: %f" % (cap)
		return cap 
	
	def removeMasterVM(self, vm):
		assert(vm.volume == 0 and vm.isMaster())
		for v in vm.instances:
			v.machine.__virtualMachines[v.function.name].remove(v)
	
	@staticmethod
	def canSwap(vms1, vms2):
		assert(vms1 and vms2)
		assert(all(v1.machine == v2.machine for v1 in vms1 for v2 in vms1))
		assert(all(v1.machine == v2.machine for v1 in vms2 for v2 in vms2))
		assert(vms1[0].machine != vms2[0].machine)
		m1 = vms1[0].machine
		m2 = vms2[0].machine
		
		vms1S = set(vms1)
		vms1After = [v for v in m1.virtualMachines if v not in vms1S]
		vms2S = set(vms2)
		vms2After = [v for v in m2.virtualMachines if v not in vms2S]
		
		vms1V = reduce(lambda s, x: s + x.volume, vms1, 0)
		vms2V = reduce(lambda s, x: s + x.volume, vms2, 0)
		return (
	(m1.residualCapacity() >= vms2V - vms1V) and 
	(m2.residualCapacity() >= vms1V - vms2V) and 
	not any(vm in vms2After for vm1 in vms1 for vm in vm1.instances) and
	not any(vm in vms1After for vm2 in vms2 for vm in vm2.instances) and
	not any(vm.isMaster() and filter(lambda v: v.isMaster() and v.function == vm.function, vms1After) for vm in vms2) and 
	not any(vm.isMaster() and filter(lambda v: v.isMaster() and v.function == vm.function, vms2After) for vm in vms1))
	
	@staticmethod
	def swap(vms1, vms2):
		assert(vms1 and vms2)
		assert(all(v1.machine == v2.machine for v1 in vms1 for v2 in vms1))
		assert(all(v1.machine == v2.machine for v1 in vms2 for v2 in vms2))
		assert(vms1[0].machine != vms2[0].machine)
		m1 = vms1[0].machine
		m2 = vms2[0].machine
		
		assert(all(vm in m1.__virtualMachines[vm.function.name] for vm in vms1))
		assert(all(vm in m2.__virtualMachines[vm.function.name] for vm in vms2))
		
		vms1V = reduce(lambda s, x: s + x.volume, vms1, 0)
		vms2V = reduce(lambda s, x: s + x.volume, vms2, 0)
		
		#assert(Machine.canSwap(vm1, vm2))
		assert(m1.residualCapacity() >= vms2V - vms1V )
		assert(m2.residualCapacity() >= vms1V - vms2V )
		for vm in vms1:
			m1.__virtualMachines[vm.function.name].remove(vm)
		for vm in vms2:
			m2.__virtualMachines[vm.function.name].remove(vm)
			
		for vm in vms1:
			assert(not vm.isMaster() or not m2.hasMasterVirtualMachine(vm.function))
			vm.move(m2)
			m2.__virtualMachines.setdefault(vm.function.name,[]).append(vm)
		for vm in vms2:
			assert(not vm.isMaster() or not m1.hasMasterVirtualMachine(vm.function))
			vm.move(m1)
			m1.__virtualMachines.setdefault(vm.function.name,[]).append(vm)
		
	def removeVM(self, vm):
		self.__virtualMachines[vm.function.name].remove(vm)
		vm.instances.remove(vm)
		
	def hasVMCopy(self, virtualMachine):
		self.__virtualMachines.setdefault(virtualMachine.function.name, [])
		res = any(
				(virtualMachine == vm) or
				(virtualMachine.isMaster() and virtualMachine == vm.master) or
				(vm.isMaster() and virtualMachine.master == vm) or
				(not vm.isMaster() and virtualMachine.master == vm.master) 
				for vm in self.__virtualMachines[virtualMachine.function.name])
		return res
	
	def isAvailable(self, demand, function = None):
		return self.residualCapacity(function) > 0 if Machine.LOAD_BALANCING else self.residualCapacity(function) >= demand
		
	def __str__(self):
		return "("+self.__name+", a:"+str(self.__availability)+", Q:"+str(self.__capacity)+ " c:"+str(self.cluster) +")"
	
	def __repr__(self):
		return "("+self.__name+", a:"+str(self.__availability)+", Q:"+str(self.__capacity)+ " c:"+str(self.cluster) +")"

	def __deepcopy__(self, memo):
		cls = self.__class__
		newone = cls.__new__(cls)
		
		newone.__name = self.__name
		newone.__capacity = self.__capacity
		newone.cluster = None
		newone.__virtualMachines = {}
		newone.__availability = self.__availability
		
		return newone

class Cluster(object):
	
	__counter = 0
	
	def __init__(self, availability = 1.0, objid = None):
		if objid:
			assert Cluster.__counter <= int(objid)
			Cluster.__counter = int(objid)
		else:
			Cluster.__counter += 1
		self.__name = "c"+str(Cluster.__counter)
		self.__availability = float(availability)
		self.__machines = []
		self.accessLinks = {}
		self.synchronizationLinks = {}
		
	@property
	def name(self):
		return self.__name
	
	@property
	def availability(self):
		return self.__availability
	
	@property
	def machines(self):
		return self.__machines
	
	def addMachine(self, machine):
		self.__machines.append(machine)
		machine.cluster = self
	
	def __str__(self):
		return "<"+self.__name+", a:"+str(self.__availability)+">"
		
	def __repr__(self):
		return "<"+self.__name+">"
	
	def addAccessLink(self, ep, a):
		self.accessLinks[ep.name] = Link(destination = ep, availability = a)

	def addSynchronizationLink(self, c, a):
		self.synchronizationLinks[c.name] = Link(destination = c, availability = a)

	def __deepcopy__(self, memo):
		cls = self.__class__
		newone = cls.__new__(cls)
		newone.__name = self.__name
		newone.__availability = self.__availability
		newone.__machines = []
		newone.accessLinks = self.accessLinks
		newone.synchronizationLinks = self.synchronizationLinks
		
		return newone
	
	
if __name__ == '__main__':
	from .functions import Function
	from .request import Request
	
	clusters = [Cluster() for _ in range(0,3)]

	machines = [Machine(capacity = 100, cluster = clusters[0]) for _ in range(0,3)]
	
	r = Request(function = Function(), demand = 10 )
	
	print(machines)
	
	print (clusters)
