from lib.cluster import computeAvailability

from copy import deepcopy
import pickle
from pulp import Amply
from lib.request import Request
from lib.functions import Function
from lib.cluster import Cluster, Machine
from lib.entrypoints import EntryPoint

def readInstance(path):
    return readAmplInstance(path) if "ampl.dat" in path else readBinaryInstance(path)

def readSolution(path, instance):
    return readAmplSolution(path, instance) # if "ampl.dat" in path else readBinarySolution(path)

def readBinaryInstance(path):
    return pickle.load(open(path,"rb"))
    
def readAmplInstance(path):
    instance = Instance()
    
    data = Amply()
    
    with open(path, "r") as f:
        data.load_file(f)

    instance.availability = data.SLA
    
    # Create clusters
    cluster_dict = {k : Cluster(a, k) for k, a in data.availability_c.data.iteritems()}
    instance.clusters = [f for k, f in cluster_dict.iteritems() ]
    
    # Create servers
    machine_dict = {k : Machine(availability = a, capacity = data.q.data[k][1], objid = k) for k, a in data.availability_s.data.iteritems()}
    instance.machines = [f for k, f in machine_dict.iteritems() ]
    
    # Add machines to clusters
    for k, m in machine_dict.iteritems():
        cluster_dict[data.server_cluster[k]].addMachine(m)
        
    # Create functions
    function_dict = {k : Function(a, objid = k) for k, a in data.availability_f.data.iteritems()}
    instance.functions = [f for k, f in function_dict.iteritems() ]
    
    # Add link between clusters
    for s, ls in data.availability_link_cc.data.iteritems():
        for d, a in ls.iteritems():
            c1 = cluster_dict[s]
            c2 = cluster_dict[d]
            if c1 != c2:
                c1.addSynchronizationLink(c2, a)

    # Create access points and add links to clusters
    ap_dict = {}
    for s, ls in data.availability_link_ac.data.iteritems():
        ap = ap_dict.setdefault(s, EntryPoint(objid = s))
        for d, a in ls.iteritems():
            c = cluster_dict[d]
            c.addAccessLink(ap, a)
            ap.addAccessLink(c, a)
    instance.entryPoints = [ep for _, ep in ap_dict.iteritems()]
                
    # Create requests
    request_dict = {int(r) : [function_dict[f], data.penalty_r.data[r], data.d.data[r][1]] for r, f in data.function_r.data.iteritems()}
    for r, ds in data.apoints.data.iteritems():
        request_dict[r].append([ ap_dict[e] for e, b in ds.iteritems() if int(b) == 1 ])
    instance.requests = [Request(eps, f, d, p, objid = r) for r, [f, p, d, eps] in request_dict.iteritems()]   
    
    return instance

def readAmplSolution(path, instance):
    data = Amply()
    
    with open(path, "r") as f:
        data.load_file(f)
    
    for m in instance.machines:
        m_id = int(m.name.replace("m","",1))
        for r in instance.requests:
            r_id = int(r.name.replace("r","",1))
            d = data.assignments.data[m_id][r_id]
            if d > 0:
                m.addRequest(r,d)

    for m_master in instance.machines:
        for m_slave in instance.machines:
            for f in instance.functions:
                mm_id = int(m_master.name.replace("m","",1))
                ms_id = int(m_slave.name.replace("m","",1))
                f_id = int(f.name.replace("f","",1))
                
                if data.slaves.data[mm_id][ms_id][f_id] == 1:
                    m_master.masterVirtualMachine(f).backup(m_slave)
                    
class Instance:
    
    def __init__(self):
        self.availability = 1.0
        self.requests = []
        self.machines = []
        self.clusters = []
        self.entryPoints = []
        self.functions = []
        self.objective = "Penalty"
        self.__cache = False
        self.__assigned = 0
        self.__penalty = 1e100
        self.__availability = 0
    
    def servedRequests(self):
        return sum(1 if r.isAssigned() else 0 for r in self.requests)
    
    def getQuality(self, cache = False):
        if cache:
            assert self.__cache
            return self.__penalty if self.objective == "Penalty" else self.__availability
        
        self.__cache = True
        
        max_penalty = 0.0
        min_availability = 1.0
        assignedRequests = len(filter(lambda r: r.isAssigned(), self.requests))
        
        cache = {}
        
        for request in self.requests:
            if request.isAssigned():
                availability = 1.0
                for masterVM in request.masters:
                    availability *= cache.setdefault(str(request.entryPoints)+"_"+str(masterVM), computeAvailability(request.entryPoints, masterVM))
            else:
                availability = 0.0
            
            max_penalty = max(max_penalty, float(self.availability - availability) * float(request.penalty))
            min_availability = min(min_availability, availability)
        
        self.__assigned = assignedRequests
        self.__penalty = max_penalty
        self.__availability = min_availability
        
        return self.__penalty if self.objective == "Penalty" else self.__availability
    
    def computeQuality(self, request):
        availability = 1.0 if request.masters else 0.0
        for masterVM in request.masters:
            availability *= computeAvailability(request.entryPoints, masterVM)
        
        max_penalty = max(0.0, float(self.availability - availability) * float(request.penalty))
        min_availability = min(1.0, availability)
        return max_penalty if self.objective == "Penalty" else min_availability
    
    def getWorstRequests(self):
        qualities = { r : self.computeQuality(r) for r in self.requests }
        max_val = max(qualities.values())
        return [ r for r,v in qualities.iteritems() if v == max_val]

    def isAtBest(self):
        val = self.getQuality()
        return val == 0.0 if self.objective == "Penalty" else val == 1.0

    def info(self):
        return "#Requests: %d\n#Servers: %d\n#Clusters: %d\nSLA: %.6f\
        \n#Demand: %f\nCapacity: %f" % (len(self.requests), len(self.machines), len(self.clusters), self.availability,
                                        reduce(lambda s, r: s + r.demand, self.requests, 0), reduce(lambda s, m: s + m.capacity, self.machines, 0))
        

    def stats(self):
        self.getQuality()
        return "Assigned: %d/%d\nPenalty: %.8f\nAvailability: %.8f" % (self.__assigned, len(self.requests), self.__penalty, self.__availability)
    
    def isBetterThan(self, instance, selfRecompute = True, otherRecompute = True):
        assert(self.objective == instance.objective)
        q1 = self.getQuality(not selfRecompute)
        q2 = instance.getQuality(not otherRecompute)
        wr1 = len(self.getWorstRequests())
        wr2 = len(instance.getWorstRequests())
        return (q1 < q2 if self.objective == "Penalty" else q1 > q2) or (q1 == q2 and wr1 < wr2)


    def saveInstance(self, path, saveType):
        if saveType == "BINARY":
            self.saveBinaryInstance(path+".bin.dat")
        elif saveType == "TEXT":
            self.saveTextInstance(path)
        else:
            raise Exception("Invalid type %s" % (saveType))
            
    def saveBinaryInstance(self, path):
        with open(path, 'wb') as output:
            pickle.dump(self, output)

    def saveTextInstance(self, path):
        with open(path, 'w') as output:
            output.write("param REQUESTS := %d;\n" % (len(self.requests)))
            output.write("param SERVERS := %d;\n" % (len(self.machines)))
            output.write("param CLUSTERS := %d;\n" % (len(self.clusters)))
            output.write("param RESOURCES := %d;\n" % (1))
            output.write("param FUNCTIONS := %d;\n" % (len(self.functions)))
            output.write("param ACCESS_POINTS := %d;\n\n" % (len(self.entryPoints)))
            output.write("param SLA := %f;\n\n" % (self.availability))
            
            output.write("param availability_f{F};\nparam availability_f := \n%s\n;\n\n" 
                         % ("\n".join(["%d %f" % (int(f.name.replace("f","",1)), f.availability) for f in self.functions])))
            
            output.write("param availability_s{S};\nparam availability_s := \n%s\n;\n\n" 
                         % ("\n".join(["%d %f" % (int(m.name.replace("m","",1)), m.availability) for m in self.machines])))
            
            output.write("param availability_c{C};\nparam availability_c := \n%s\n;\n\n" 
                         % ("\n".join(["%d %f" % (int(c.name.replace("c","",1)), c.availability) for c in self.clusters])))
    
            output.write("param availability_link_cc{C,C};\nparam availability_link_cc : %s := \n%s\n;\n\n" 
                         % (" ".join(c.name.replace("c","",1) for c in self.clusters),
                            "\n".join("%d %s" % (int(c1.name.replace("c","",1)), 
                                                  " ".join(str(c1.synchronizationLinks[c2.name].availability if c1 != c2 else 1.0) for c2 in self.clusters)) for c1 in self.clusters)))
            
            output.write("param availability_link_ac{A,C};\nparam availability_link_ac : %s := \n%s\n;\n\n" 
                         % (" ".join(c.name.replace("c","",1) for c in self.clusters),
                            "\n".join("%d %s" % (int(e.name.replace("ep","",1)), 
                                                  " ".join(str(e.accessLinks[c.name].availability) for c in self.clusters)) for e in self.entryPoints)))

            output.write("param penalty_r{R};\nparam function_r{R};\nparam : penalty_r function_r := \n%s\n;\n\n" 
                         % ("\n".join(["%d %f %d" % (int(r.name.replace("r","",1)), r.penalty, int(r.function.name.replace("f","",1))) for r in self.requests])))

            output.write("param apoints{R,A};\nparam apoints : %s := \n%s\n;\n\n" 
                         % (" ".join(e.name.replace("ep","",1) for e in self.entryPoints),
                            "\n".join(["%d %s" % (int(r.name.replace("r","",1)), 
                                                  " ".join("1" if e in r.entryPoints else "0" for e in self.entryPoints)) for r in self.requests])))

            output.write("param d{R,K};\nparam d : 1 := \n%s\n;\n\n" 
                         % ("\n".join(["%d %f" % (int(r.name.replace("r","",1)), r.demand) for r in self.requests])))

            output.write("param q{S,K};\nparam q : 1 := \n%s\n;\n\n" 
                         % ("\n".join(["%d %f" % (int(m.name.replace("m","",1)), m.capacity) for m in self.machines])))
            
            output.write("param server_cluster{S};\nparam server_cluster := \n%s\n;\n\n" 
                         % ("\n".join(["%d %d" % (int(m.name.replace("m","",1)), int(m.cluster.name.replace("c","",1))) for m in self.machines])))

    def savePlacement(self, path, saveType):
        if saveType == "BINARY":
            self.saveBinaryInstance(path+".bin.dat")
        elif saveType == "TEXT":
            self.saveTextPlacement(path)
        else:
            raise Exception("Invalid type %s" % (saveType))

    def saveTextPlacement(self, path):
        with open(path, 'w') as output:
            output.write("param vnfs{S,F};\nparam vnfs : %s := \n" % (" ".join(map(lambda f: f.name.replace("f", "", 1), self.functions))))
            for m in self.machines:
                output.write("%s %s\n" % (m.name.replace("m","",1), 
                                        " ".join([str(m.masterVirtualMachine(f).volume) if m.hasMasterVirtualMachine(f) else str(0.0) for f in self.functions]) ))
            output.write(";\n")

            output.write("param slaves{S,S,F};\nparam slaves := \n")
            for f in self.functions:
                output.write("[*,*,%s]: %s :=\n" % (f.name.replace("f","",1), " ".join(map(lambda m: m.name.replace("m", "", 1), self.machines))))
                for m in self.machines:
                    output.write("%s %s\n" % (m.name.replace("m","",1), 
                                              " ".join(["1" if m != m2 and m.hasMasterVirtualMachine(f) and any(vm.machine == m2 for vm in m.masterVirtualMachine(f).instances) 
                                                        else "0" for m2 in self.machines])))
            output.write(";\n")
            
            output.write("param assignments{S,R};\nparam assignments : %s := \n" % (" ".join(map(lambda r: r.name.replace("r", "", 1), self.requests))))
            for m in self.machines:
                output.write("%s %s\n" % (m.name.replace("m","",1), 
                                        " ".join([str(m.assignment(r)) if m.hasRequest(r) else str(0.0) for r in self.requests]) ))
            output.write(";\n")
            
            self.getQuality()
                        
            output.write("param assigned_requests := %d;\n" % (self.__assigned))
            output.write("param max_availability := %.6f;\n" % (self.__availability))
            output.write("param min_penalty := %.6f;\n" % (self.__penalty))
        
    def __deepcopy__(self, memo):
        newone = Instance()
        
        newone.availability = self.availability
        newone.requests = deepcopy(self.requests)
        newone.machines = deepcopy(self.machines)
        newone.clusters = deepcopy(self.clusters)
        newone.entryPoints = self.entryPoints
        newone.functions = self.functions
        newone.objective = self.objective
        newone.__cache = self.__cache
        newone.__assigned = self.__assigned
        newone.__penalty = self.__penalty
        newone.__availability = self.__availability
        mapNewClusters = {c.name : c for c in newone.clusters}
        mapNewMachines = {m.name : m for m in newone.machines}
        mapNewRequests = {r.name : r for r in newone.requests}
        
        for c in self.clusters:
            for m in c.machines:
                newMachine = mapNewMachines[m.name]
                mapNewClusters[c.name].addMachine(newMachine)
                for r in m.requests:
                    newMachine.addRequest(mapNewRequests[r.name], m.assignment(r))                    

                assert all(newMachine.masterVirtualMachine(vm.function) != None for vm in m.virtualMachines if vm.isMaster() and vm.volume > 0)
        
                for vm in m.virtualMachines:
                    if vm.isMaster() and vm.volume > 0:
                        for backup in vm.instances:
                            if backup != vm:
                                newMachine.masterVirtualMachine(vm.function).backup(mapNewMachines[backup.machine.name])
        
             
        
        assert all(m.residualCapacity() == mapNewMachines[m.name].residualCapacity() for c in self.clusters for m in c.machines)
        return newone