#!/usr/bin/env python

from functools import reduce
import pickle
import sys
import greedy_heuristic
from lib.cluster import Machine
from copy import deepcopy
from lib.timer import Timer
from lib.instance import readInstance

timer = Timer()


def splitMasterStructure(instance, request, wr):
    bestCopy = deepcopy(instance)
    update = False
    for vm in request.masters:
        for backup in [b for b in vm.instances if b != vm and not b.machine.hasMasterVirtualMachine(request.function)]: 
            d = vm.assignment(request)
            vm.removeRequest(request, d)
            backup.machine.addRequest(request,d)
            newMaster = backup.machine.masterVirtualMachine(request.function) 
            addBackup(instance, newMaster)
            if instance.isBetterThan(bestCopy):
                bestCopy = deepcopy(instance)
                update = True
                return bestCopy
            
            newMaster.removeRequest(request, d)
            backup.machine.removeMasterVM(newMaster)
            vm.machine.addRequest(request,d)
                
    return bestCopy if update else None

def changeBackupStructure(instance, request, wr):
    bestCopy = deepcopy(instance)
    update = False
    for vm in request.masters:
        for m in (m for m in instance.machines if all(v.machine != m for v in vm.instances)):
            for backup in (b for b in m.virtualMachines if not b.isMaster()):
                if m.residualCapacity() + backup.volume - vm.volume >= 0 :
                    masterB = backup.master
                    m.removeVM(backup)
                    assert masterB
                    
                    newBackup = vm.backup(m)
                    
                    if instance.isBetterThan(bestCopy):
                        bestCopy = deepcopy(instance)
                        update = True
                        return bestCopy
                    
                    m.removeVM(newBackup)
                    masterB.backup(m)
                    
    return bestCopy if update else None

def masterChangeStructure(instance, request, wr):
    bestCopy = deepcopy(instance)
    update = False
    for vm1 in request.masters:
        for vm2 in vm1.instances:
            if vm1 != vm2:
                vm1.changeMaster(vm2)
                if instance.isBetterThan(bestCopy):
                    bestCopy = deepcopy(instance)
                    update = True
                    return bestCopy
                
                vm2.changeMaster(vm1)
    return bestCopy if update else None

def requestMoveStructure(instance, request, wr):
    bestCopy = deepcopy(instance)
    update = False
    for masterVM in request.masters: 
        for machine in [m for m in instance.machines if m != masterVM.machine]:
            d = masterVM.assignment(request)
            masterVM.removeRequest(request,d)
            
            vm2 = machine.masterVirtualMachine(request.function)
            if (vm2 == None and machine.residualCapacity() >= d) or (vm2 and all(vm.machine.residualCapacity() >= d for vm in vm2.instances)):
                machine.addRequest(request,d)
                vm = machine.masterVirtualMachine(request.function)
                backups = addBackup(instance, vm)
                if instance.isBetterThan(bestCopy):
                    bestCopy = deepcopy(instance)
                    update = True
                    return bestCopy
                
                for v in backups:
                    v.machine.removeVM(v)
                vm.removeRequest(request,d)
            
            masterVM.machine.addRequest(request,d)
    return bestCopy if update else None

POWERSET_MAX_SIZE = 1

def limited_powerset(seq, limit, l_func):
    res = [[]]
    seq.sort(key=l_func, reverse=True )
    for s in seq:
        temp = []
        for r in res:
            if len(r) < POWERSET_MAX_SIZE and reduce(lambda s, x: s + l_func(x), r, 0) < limit:
                temp.append([s] + r)
        res += temp
    return res

def requestSwapStructure(instance, request, wr):
    bestCopy = deepcopy(instance)
    update = False
    for masterVM1 in request.masters:
        for m in (m for m in instance.machines if m != masterVM1.machine):
            for rs in limited_powerset(m.requests, request.demand, lambda x : x.demand): #[[r] for r in m.requests]:
                d1 = masterVM1.assignment(request)
                masterVM1.removeRequest(request, d1)
                rs = [[r,m.assignment(r)] for r in rs]
                for [r,d] in rs:
                    m.removeRequest(r, d)
                addR = []
                
                for [r, d] in rs:
                    vm2 = masterVM1.machine.masterVirtualMachine(r.function)
                    if vm2 == None and masterVM1.machine.residualCapacity() >= d:
                        addR.append([r, d])
                        masterVM1.machine.addRequest(r, d)
                    elif vm2 and all(vm.machine.residualCapacity() >= d for vm in vm2.instances):
                        addR.append([r, d])
                        masterVM1.machine.addRequest(r, d)
                    else:
                        break
                vm2 = m.masterVirtualMachine(request.function)
                if len(addR) == len(rs) and ((not vm2 and m.residualCapacity() >= d1) or 
                                             (vm2 and all(vm.machine.residualCapacity() >= d1 for vm in vm2.instances))):
                    m.addRequest(request, d1)
                    newBackups = addBackup(instance, m.masterVirtualMachine(request.function))
                    if instance.isBetterThan(bestCopy):
                        bestCopy = deepcopy(instance)
                        update = True
                        return bestCopy

                    for v2 in newBackups:
                        v2.machine.removeVM(v2)
                    m.removeRequest(request, d1)
                for [r,d] in addR:
                    masterVM1.machine.removeRequest(r, d)
                for [r,d] in rs:
                    m.addRequest(r, d)
                masterVM1.machine.addRequest(request, d1)
                
    return bestCopy if update else None

def vmSwapStructure(instance, request, wr):
    bestCopy = deepcopy(instance)
    update = False
    for vm in (v for master in request.masters for v in master.instances):
        for m in instance.machines:
            if vm.machine != m:
                for v in (v for v in limited_powerset(m.virtualMachines, vm.volume, lambda x : x.volume) if v and Machine.canSwap([vm], v)):
                    Machine.swap([vm], v)
                    newBackups = []
                    for b in [vm] + v:
                        newBackups += addBackup(instance, b)
                    if instance.isBetterThan(bestCopy):
                        bestCopy = deepcopy(instance)
                        update = True
                        return bestCopy
                    
                    for v2 in newBackups:
                        v2.machine.removeVM(v2)
                    
                    Machine.swap(v, [vm])
    
    return bestCopy if update else None
    
PROTECTION = False
def addBackup(instance, vm):
    if not PROTECTION:
        return []
    
    newBackups = []
    bestInstance = deepcopy(instance)
    
    while True:
        backup = greedy_heuristic.addProtection(vm if vm.isMaster() else vm.master, instance.machines)
        if backup:
            newBackups.append(backup)
            if instance.isBetterThan(bestInstance):
                break
        else:
            break

    return newBackups

def clean(instance):
    for m in instance.machines:
        for vm in m.virtualMachines:
            if vm.volume == 0:
                m.removeVM(vm)

def solve(sourceInstance, loadBalancing = True, protection = False):
    
    starting = [
                # 1 - 4
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, True, "BEST_AVAILABILITY"),
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, True, "BEST_FIT"),
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, True, "WORST_FIT"),
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, True, "RANDOM"),
                # 5 - 8
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, False, "BEST_AVAILABILITY"),
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, False, "BEST_FIT"),
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, False, "WORST_FIT"),
                greedy_heuristic.solve(sourceInstance, loadBalancing, protection, False, "RANDOM")
                ]
    
    Machine.VM_SIZE_FACTOR = 1.0
    PROTECTION = protection
    
    structures = [vmSwapStructure, changeBackupStructure, requestSwapStructure, requestMoveStructure]


    bestInstance = starting[0]
    #timer.setTimeLimit(600)
    for index, instance in enumerate(starting):
        
        timer.reset()
        
        if timer.checkTimeLimit():
            break
        
        structureImprove = True
        instance.getQuality()
        oldInstance = deepcopy(instance)
        
        print("Starting solution %d from %f #requests %d" % (index+1, instance.getQuality(False), len(instance.getWorstRequests())))
        
        while structureImprove and not instance.isAtBest():
            
            if timer.checkTimeLimit():
                break
        
            structureImprove = False
            for s in structures:
                improve = True
                
                wrs = instance.getWorstRequests()
                for worstRequest in wrs:
                    if timer.checkTimeLimit():
                        break
                                        
                    assert(instance.servedRequests() == len(instance.requests))
                    improve = s(instance, worstRequest, wrs)
                    clean(instance)
                    if improve:
                        clean(improve)
                        instance = improve
                        structureImprove = True
                        assert instance.isBetterThan(oldInstance), "instance: %f %d, old instance: %f %d" % (instance.getQuality(False), len(instance.getWorstRequests()), oldInstance.getQuality(False), len(oldInstance.getWorstRequests()))
                        oldInstance = deepcopy(instance)
                        break
                    
                    if instance.isBetterThan(bestInstance):
                        bestInstance = instance
                    
                print("\tStructure %s -> %s" % (s.__name__, "improving %f #requests %d" % (instance.getQuality(False), len(instance.getWorstRequests())) if improve else "not improving"))
                if improve:
                    break
        
        print("Ending with %f #requests %d" % (instance.getQuality(False), len(instance.getWorstRequests())))
        if instance.isBetterThan(bestInstance):
            bestInstance = instance
            
        if bestInstance.isAtBest():
            print("Skip following restarting points")
            return bestInstance
    
    return bestInstance

if __name__ == '__main__':
    #instance = generatore.generate(1,False)
    instance = readInstance(sys.argv[1])
#     instance = pickle.load(open(sys.argv[1],"rb"))
    
    #timer.setTimeLimit(1)
    
    balancing = sys.argv[2] == "True"
    protection = sys.argv[3] == "True"
    obj = sys.argv[4]
    instance.objective = obj

    #if len(sys.argv) > 5:
    #    timer.setTimeLimit(int(sys.argv[5]))
    
    instance = solve(instance, balancing, protection)
    
    print(instance.stats())
    print("TotelTime %.6f" % (timer.computingTime()))

    if len(sys.argv) == 6:
        instance.savePlacement(sys.argv[5], "TEXT")
